"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

import inspect
import traceback
import types
import importlib
import sys

import Octrapy
from .core import Command
from .cog import Cog
from .context import Context
from .errors import *
from .help import HelpCommand, DefaultHelpCommand


_default_help = DefaultHelpCommand()


class Bot(Octrapy.Client):
    """
    Represents a Octra bot.

    Parameters
    ----------
    token: :class:`str`
        The API token.
    description: Optional[:class:`str`]
        The bot's description.
    owner_id: Optional[:class:`int`]
        The owner's ID.
    owner_ids: Optional[List[:class:`int`]]
        The owner IDs.
    help_command: Optional[:class:`Octrapy.ext.commands.HelpCommand`]
        The bot's help command.
        Defaults to :class:`Octrapy.ext.commands.DefaultHelpCommand`

    Attributes
    ----------
    description: Optional[:class:`str`]
        The bot's description.
    owner_id: Optional[:class:`int`]
        The owner's ID.
    owner_ids: Optional[List[:class:`int`]]
        The owner IDs.
    cogs: Mapping[:class:`str`: :class:`Octrapy.ext.commands.Cog`]
        A dictonary of cogs that are loaded.
    extensions: Mapping[:class:`str`: :class:`types.ModuleType`]
        A dictonary of extensions that are loaded.
    """

    def __init__(self, token: str,  *, description: str = None, owner_id: int = None, owner_ids: list = None, help_command: HelpCommand = _default_help):
        super().__init__(token)
        self.owner_id = owner_id
        self.owner_ids = owner_ids
        self.description = description

        self.commands_dict = {}

        self.cogs = {}
        self.extension_cogs = {}
        self.extensions = {}

        if help_command:
            self.help_command = help_command

    @property
    def help_command(self):
        """:class:`Octrapy.ext.commands.HelpCommand`:
            The bot's help command.
        """
        return self._help_command

    @help_command.setter
    def help_command(self, value: HelpCommand):
        if not isinstance(value, HelpCommand):
            raise TypeError("The new help command must inherit from HelpCommand.")

        value._add_to_bot(self)
        self._help_command = value

    @property
    def commands(self):
        """
        :class:`list`:
            A list of the commands.
        """

        return list(self.commands_dict.values())

    def _get_all_command_names(self):
        all_names = []
        for command in self.commands_dict.values():
            all_names.append(command.name)
            all_names.extend(command.aliases)

        return all_names

    def get_command(self, name: str):
        """
        Gets a command by name.

        Parameters
        ----------
        name: :class:`str`
            The command name.

        Returns
        --------
        :class:`Octrapy.ext.commands.Command`
            The command with the name.
        """

        for command in self.commands_dict.values():
            if name in command.aliases or command.name == name:
                return command

    async def get_context(self, message: Octrapy.Message):
        """
        Gets context for a given message.

        Parameters
        ----------
        message: :class:`Octrapy.Message`
            The message to get context from.

        Returns
        -------
        :class:`Octrapy.ext.commands.Context`
            The context created.

        Raises
        ------
        :class:`Octrapy.ext.commands.CommandNotFound`
            The command specified was not found.
        """

        content = message.content.split(" ")[0]
        command = None
        if content.startswith("/"):
            splited = content.split("@")
            if len(splited) == 1 or (len(splited) != 1 and splited[1] == (await self.user()).username):
                command_name = splited[0][1:]
                command = self.get_command(command_name)
                if not command:
                    raise CommandNotFound(command_name)

        kwargs = {"command": command}
        kwargs["args"] = []
        kwargs["kwargs"] = {}
        kwargs["message"] = message
        kwargs["chat"] = message.chat
        kwargs["author"] = message.author
        kwargs["bot"] = self
        return Context(**kwargs)

    def load_extension(self, location: str):
        """
        Loads an extension.

        Parameters
        ----------
        location: :class:`str`
            The location of the extension.

        Raises
        ------
        :exc:`Octrapy.ExtensionAlreadyLoaded`
            The extension is already loaded.
        :exc:`AttributeError`
            The extension has no setup function.
        """

        if location in self.extensions:
            raise ExtensionAlreadyLoaded(location)

        cog = importlib.import_module(location)
        self.extensions[cog.__name__] = cog

        if not hasattr(cog, "setup"):
            raise AttributeError("Extension has no setup function")

        cog.setup(self)

    def unload_extension(self, location: str):
        """
        Unloads an extension.

        Parameters
        ----------
        location: :class:`str`
             The location of the extension.
        """

        cog_name = self.extension_cogs.get(location)
        if not cog_name:
            raise ExtensionNotLoaded(location)

        self.remove_cog(cog_name)

        self.extension_cogs.pop(location)
        self.extensions.pop(location)

    def reload_extension(self, location: str):
        """
        Reloads an extension.

        Parameters
        ----------
        location: :class:`str`
            The location of the extension.
        """

        if location not in self.extension_cogs:
            raise ExtensionNotLoaded(location)

        lib = sys.modules[location]
        importlib.reload(lib)
        self.remove_cog(self.extension_cogs[location])

        sys.modules[location].setup(self)

    def add_cog(self, cog: Cog):
        """
        Adds a cog to the bot.

        Parameters
        ----------
        cog: :class:`Octrapy.ext.commands.Cog`
            The cog to add.

        Raises
        ------
        :exc:`TypeError`
            The cog is not a subclass of :class:`Octrapy.ext.commands.Cog` or the cog check is not a method.
        """

        if not isinstance(cog, Cog):
            raise TypeError("Cog is not a subclass of Cog")

        #Add the cog
        cog._add(self)
        self.cogs[cog.qualified_name] = cog
        # If cog is from an extension, add the extension to a dict with the cog name as the value
        if str(cog.__module__) != "__main__":
            self.extension_cogs[cog.__module__] = cog.qualified_name

    def remove_cog(self, cog: str):
        """
        Removes and cog from the bot.

        Parameters
        ----------
        cog: :class:`str`
            The name of the cog to remove.
        """

        cog = self.cogs.get(cog)
        if not cog:
            return
        cog._remove(self)
        self.cogs.pop(cog.qualified_name)

    async def _use_command(self, ctx):
        await self._dispatch("command", ctx)

        try:
            await ctx.command.invoke(ctx)
        except Exception as exc:
            ctx.command_failed = True
            await self._dispatch("command_error", ctx, exc)

        if not ctx.command_failed:
            await self._dispatch("command_completion", ctx)

    async def _dispatch(self, event, *args):
        await super()._dispatch(event, *args)

        # If event is on_message, check if the message if a command
        # If it is a command, invoke the command
        if event == "message":
            message = args[0]
            if not message.content:
                return

            ctx = await self.get_context(message)

            if not ctx.command:
                return

            self.loop.create_task(self._use_command(ctx))

    def command(self, *args, **kwargs):
        """
        Turns a function into a command.

        Parameters
        ----------
        \*args:
            The arguments.
        \*\*kwargs:
            The keyword arguments.
        """

        def deco(func):
            name = kwargs.get("name") or func.__name__

            kwargs["name"] = name
            if name in self._get_all_command_names():
                raise CommandRegistrationError(name)

            command = Command(func, **kwargs)
            command.checks = getattr(func, "_command_checks", [])

            self.commands_dict[name] = command
            return command

        return deco

    def add_command(self, command):
        """
        Adds a command.

        Parameters
        ----------
        command: :class:`Octrapy.ext.commands.Command`
            The command to add.

        Returns
        -------
        :class:`Octrapy.ext.commands.Command`
            The command added.

        Raises
        ------
        :exc:`Octrapy.ext.commands.CommandRegistrationError`
            The command name or one if its aliases is already registered by a different command.
        :exc:`TypeError`
            The command is not an instance of :class:`Octrapy.ext.commands.Command`.
        """

        if not isinstance(command, Command):
            raise TypeError("Command must be a subclass of Command")

        if command.name in self._get_all_command_names():
            raise CommandRegistrationError(command.name)

        self.commands_dict[command.name] = command
        return command

    def remove_command(self, name: str):
        """
        Removes a command by name.

        Parameters
        ----------
        name: :class:`str`
            The name of the command to remove.

        Returns
        -------
        :class:`Octrapy.ext.commands.Command`:
            The command removed.
        """

        command = self.get_command(name)

        if not command:
            return

        self.commands_dict.pop(name)

        return command

    async def on_command_error(self, ctx, error):
        """The default command error handler."""

        if self._listeners.get("on_command_error"):
            return

        print(f"Ignoring exception in command {ctx.command}:", file=sys.stderr)
        traceback.print_exception(
            type(error), error, error.__traceback__, file=sys.stderr
        )



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

class OctraException(Exception):
    """Base exception for all errors."""

    pass

class HTTPException(OctraException):
    """
    Raised when an HTTP request fails.

    Attributes
    ----------
    response: :class:`aiohttp.ClientResponse`
        The response for the request that failed.
    messsage: Optional[:class:`str`]
       The message for the request that failed.
    """

    def __init__(self, response, message=""):
        self.response = response
        self.message = message
        super().__init__(f"{response.status} {message}")

class InvalidToken(HTTPException):
    """Raised when a token is invalid."""

    def __init__(self, response, message=""):
        super().__init__(response, message)

class Forbidden(HTTPException):
    """Raised when something is forbidden."""

    def __init__(self, response, message=""):
        super().__init__(response, message)

class Conflict(HTTPException):
    """Raised when another instance of the bot is running."""

    def __init__(self, response, message=""):
        super().__init__(response, message)


"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

from Octrapy import File


class Context:
    """
    Context for a command.

    Attributes
    ----------
    bot: :class:`Octrapy.ext.commands.Bot`
        The bot that created the context.
    message: :class:`Octrapy.Message`
        The message the context is for.
    chat: :class:`Octrapy.Chat`
        The chat the context is for.
    author: :class:`Octrapy.User`
        The author of the message.
    command_failed: :class:`bool`
        Whether the command failed or not.
    args: :class:`list`
        The arguments passed into the command.
    kwargs: :class:`dict`
        The kwargs passed into the command.
    """

    def __init__(self, command, **kwargs):
        self.command = command
        self.bot = kwargs.get("bot")

        self.message = kwargs.get("message")
        self.chat = kwargs.get("chat")
        self.author = kwargs.get("author")
        self.command_failed = None
        self.args = kwargs.get("args") or []
        self.kwargs = kwargs.get("kwargs") or {}

    async def send(self, content: str = None, file: File = None, parse_mode=None):
        """|coro|
        
        Sends a message in the chat.

        Parameters
        ----------
        content: :class:`str`
            The content of the message to send.
        file: :class:`Octrapy.File`
            The file to send.
        parse_mode: :class:`str`
            The parse mode of the message to send.

        Returns
        -------
        :class:`Octrapy.Message`
            The message sent.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Sending the message failed.
        """

        return await self.chat.send(content=content, file=file, parse_mode=parse_mode)

    async def send_poll(self, question: str, options: list):
        """|coro|

        Sends a poll to the chat.

        Parameters
        ----------
        question: :class:`str`
            The question of the poll.
        options: List[:class:`str`]
            The options in the poll.

        Returns
        -------
        :class:`Octrapy.Poll`
            The poll sent.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Sending the poll failed.
        """

        return await self.chat.send_poll(question, options)

    async def send_action(self, action: str):
        """|coro|

        Sends an action to the chat.

        Parameters
        ----------
        action: :class:`str`
            The action to send.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Sending the action failed.
        """

        await self.chat.send_action(action)

    async def reply(self, content: str, parse_mode: str = None):
        """
        |coro|

        Replys to the message.

        Parameters
        ----------
        content: :class:`str`
            The content of the message to send.
        parse_mode: :class:`str`
            The parse mode of the message to send.

        Returns
        -------
        :class:`Octrapy.Message`
            The message sent.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Sending the message failed.
        """

        return await self.message.reply(content=content, parse_mode=parse_mode)



"""
Octra.py: Octra Bot API Wrapper
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
An async API wrapper for Octra bot API in Python.

:copyright: (c) 2023 Akash
:license: MIT, see LICENSE for more details.
"""

__title__ = "Octrapy"
__author__ = "Akash"
__license__ = "MIT"
__copyright__ = "Copyright 2023 Akash"
__version__ = "1.8.68"

from collections import namedtuple

from . import utils
from .client import Client
from .errors import *
from .chat import Chat
from .user import User
from .message import Message
from .file import *
from .poll import Poll

VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")

version_info = VersionInfo(major=0, minor=3, micro=1, releaselevel="final", serial=0)



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

import asyncio
import datetime
import json
import sys
import logging

import aiohttp

from . import __version__
from .errors import *
from .user import User
from .chat import Chat
from .message import Message
from .file import *
from .poll import *

log = logging.getLogger("Octrapy.http")

user_agent = "OctraBot (https://github.com/Octra/Octra.py {0}) Python/{1[0]}.{1[1]} aiohttp/{2}"


class Route:
    """Represents a Octra route"""

    BASE_URL = ""

    def __init__(self, method, url):
        self.method = method
        self.url = url

    def __str__(self):
        return f"{self.method}: {self.url}"


class HTTPClient:
    """Represents an HTTP client making requests to Octra."""

    def __init__(self, token: str, loop: asyncio.BaseEventLoop):
        self._token = token
        self._base_url = f"https://api.Octra.org/bot{self._token}/"
        self.messages_dict = {}

        self.loop = loop or asyncio.get_event_loop()
        self.user_agent = user_agent.format(__version__, sys.version_info, aiohttp.__version__)
        self.session = aiohttp.ClientSession(loop=self.loop, headers={"User-Agent": self.user_agent})

    @property
    def messages(self):
        """A cache of messages the bot can see."""

        return list(self.messages_dict.values())

    async def request(self, route: Route, **kwargs):
        """Make a request to a route.

        All kwargs will be forwarded to
        :meth:`aiohttp.ClientSession.request`.

        Parameters
        ----------
        route: :class:`Octrapy.http.Route`
            The route to make a request to.
        """
        url = route.url
        method = route.method

        # Try a request 5 times before dropping it
        for tries in range(5):
            async with self.session.request(method, url, **kwargs) as resp:
                # Octra docs say all responses will have json
                data = await resp.json()

                if 300 > resp.status >= 200:
                    return data

                # We are getting ratelimited
                if resp.status == 429:
                    params = data.get("parameters")

                    if not params:
                        raise HTTPException(resp, data.get("description"))

                    retry_after = params.get("retry_after")

                    # We didn't get a retry after,
                    # so raise an HTTPException
                    if not retry_after:
                        raise HTTPException(resp, data.get("description"))

                    log.warning(f"We are being ratelimited. Retrying in {retry_after} seconds.")

                    await asyncio.sleep(retry_after)
                    continue

                # Token is invalid
                if resp.status == 403:
                    raise Forbidden(resp, data.get("description"))
                if resp.status in (401, 404):
                    raise InvalidToken(resp, data.get("description"))
                if resp.status == 409:
                    raise Conflict(resp, data.get("description"))

                # Some sort of internal Octra error.
                # Retry with an increasing sleep gap between.
                if resp.status in [500, 502, 503, 504]:
                    await asyncio.sleep(1 + tries * 2)
                    continue

                else:
                    raise HTTPException(resp, (data).get("description"))

    async def send_message(self, chat_id: int, content: str, parse_mode: str = None, reply_message_id: int = None):
        """Sends a message to a chat."""

        url = self._base_url + "sendMessage"
        data = {"chat_id": chat_id, "text": content}

        if parse_mode:
            data["parse_mode"] = parse_mode
        if reply_message_id:
            data["reply_to_message_id"] = reply_message_id

        message_data = await self.request(Route("POST", url), data=data)

        if "result" in message_data:
            msg = Message(self, message_data["result"])
            self.messages_dict[msg.id] = msg
            return msg

    async def edit_message_content(self, chat_id: int, message_id: int, content: str, parse_mode: str = None):
        """Edits a message."""

        url = self._base_url + "editMessageText"
        data = {"chat_id": chat_id, "message_id": message_id, "text": content}

        if parse_mode:
            data["parse_mode"] = parse_mode

        await self.request(Route("POST", url), data=data)

    async def delete_message(self, chat_id: int, message_id: int):
        """Deletes a message."""

        url = self._base_url + "deleteMessage"
        data = {"chat_id": chat_id, "message_id": message_id}

        await self.request(Route("POST", url), data=data)

    async def forward_message(self, chat_id: int, from_chat_id: int, message_id: int):
        """Forwards a message."""

        url = self._base_url + "forwardMessage"
        data = {"chat_id": chat_id, "from_chat_id": from_chat_id, "message_id": message_id}

        message_data = await self.request(Route("POST", url), data=data)

        if "result" in message_data:
            msg = Message(self, message_data["result"])
            self.messages_dict[msg.id] = msg
            return msg

    async def send_document(self, chat_id: int, document: Document, filename: str = None):
        """Sends a document to a chat."""

        url = self._base_url + "sendDocument"

        writer = aiohttp.FormData()
        writer.add_field("document", document, filename=filename)
        writer.add_field("chat_id", str(chat_id))
        message_data = await self.request(Route("POST", url), data=writer)

        if "result" in message_data:
            msg = Message(self, message_data["result"])
            self.messages_dict[msg.id] = msg
            return msg

    async def send_photo(self, chat_id: int, photo: Photo, filename: str = None, caption: str = None):
        """Sends a photo to a chat."""

        url = self._base_url + "sendPhoto"
        writer = aiohttp.FormData()
        writer.add_field("chat_id", str(chat_id))
        writer.add_field("photo", photo, filename=filename)

        if caption:
            writer.add_field("caption", caption)

        message_data = await self.request(Route("POST", url), data=writer)

        if "result" in message_data:
            msg = Message(self, message_data["result"])
            self.messages_dict[msg.id] = msg
            return msg

    async def send_poll(self, chat_id: int, question: str, options: list):
        """Sends a poll to a chat."""

        url = self._base_url + "sendPoll"
        data = {"chat_id": chat_id, "question": question, "options": json.dumps(options)}

        poll_data = await self.request(Route("POST", url), data=data)

        if "result" in poll_data:
            msg = Poll(poll_data["result"])

    async def send_chat_action(self, chat_id: int, action: str):
        """Sends a chat action to a chat."""

        url = self._base_url + "sendChatAction"
        data = {"chat_id": chat_id, "action": action}

        chat_action_data = await self.request(Route("POST", url), data=data)

    async def get_chat(self, chat_id: int):
        """Fetches a chat."""

        url = self._base_url + "getChat"
        data = {"chat_id": chat_id}
        chat_data = await self.request(Route("GET", url), data=data)

        if "result" in chat_data:
            return Chat(self, chat_data["result"])

    async def get_chat_member(self, chat_id: int, user_id: int):
        """Fetches a member from a chat."""

        url = self._base_url + "getChatMember"
        data = {"chat_id": chat_id, "user_id": user_id}
        user_data = await self.request(Route("GET", url), data=data)

        if "result" in user_data:
            return User(self, user_data["result"].get("user"))

    async def get_me(self):
        url = self._base_url + "getMe"

        me_data = await self.request(Route("GET", url))

        if "result" in me_data:
            return User(self, me_data["result"])

    async def get_updates(self, offset=None):
        """
        Fetches the new updates for the bot.
        """

        url = self._base_url + "getUpdates"

        if not offset:
            data = await self.request(Route("GET", url))

        else:
            data = await self.request(Route("POST", url), data={"offset": offset})

        self._last_update_time = datetime.datetime.now()
        return data["result"]

    async def close(self):
        """Closes the connection."""

        await self.session.close()



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

import asyncio
import logging
import datetime
import traceback
import sys

from .http import HTTPClient
from .errors import *
from .message import Message
from .poll import Poll

log = logging.getLogger("Octrapy")


class Client:
    """
    Represents a client connection to Octra.

    Parameters
    ----------
    token: :class:`str`
        The API token.
    """

    def __init__(self, token: str):
        self._running = False
        self._last_looped = datetime.datetime.now()

        self.loop = asyncio.get_event_loop()
        self.http = HTTPClient(token=token, loop=self.loop)

        self._listeners = {}
        self._waiting_for = {}

    async def user(self):
        """|coro|

        The user of the bot.
        """

        return await self.http.get_me()

    async def get_chat(self, chat_id: int):
        """|coro|

        Fetches a chat by ID.

        Parameters
        ----------
        chat_id: :class:`int`
            The ID of the chat to fetch.

        Returns
        -------
        :class:`Octrapy.Chat`
            The chat that was fetched.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Fetching the chat failed.
        """

        return await self.http.get_chat(chat_id=chat_id)

    async def _poll(self):
        # Create some variables
        self._last_update_id = None
        self._last_update_time = datetime.datetime.now()
        self._wait_time = 1

        # Get last update id

        while True:
            try:
                log.info("Fetching unread updates")
                updates = await self.http.get_updates(self._last_update_id)
                if len(updates) != 0:
                    update_ids = [int(update["update_id"]) for update in updates]
                    self._last_update_id = max(update_ids) + 1
                break

            except InvalidToken as exc:
                traceback.print_exception(type(exc), exc, exc.__traceback__, file=sys.stderr)
                await self.stop()
                return
            except HTTPException as exc:
                traceback.print_exception(type(exc), exc, exc.__traceback__, file=sys.stderr)
                await asyncio.sleep(10)
            except Exception as exc:
                traceback.print_exception(type(exc), exc, exc.__traceback__, file=sys.stderr)
                await asyncio.sleep(10)

        # After fetching unread updates, start the loop
        while self._running:
            self._last_looped = datetime.datetime.now()

            try:
                log.debug("Fetching updates")
                updates = await self.http.get_updates(self._last_update_id)
                if len(updates) != 0:
                    log.debug(f"Handling update(s): {[update['update_id'] for update in updates]} ({len(updates)} update(s))")
                    for update in updates:
                        if "message" in update:
                            key = "message"
                            event = "message"
                        elif "edited_message" in update:
                            key = "edited_message"
                            event = "message_edit"
                        elif "poll" in update:
                            key = "poll"
                            event = "poll"

                    data = update[key]

                    if event == "poll":
                        await self._dispatch(event, Poll(data))
                    elif event == "message_edit":
                        await self._dispatch(event, self.http.messages_dict.get(data["message_id"]), Message(self.http, data))
                    else:
                        await self._dispatch(event, Message(self.http, data))

                    update_ids = [int(update["update_id"]) for update in updates]
                    self._last_update_id = max(update_ids) + 1
                    key = None
                    event = None

                for x in [x for x in updates if "message" in x]:
                    self.http.messages_dict[x["message"]["message_id"]] = Message(self.http, x["message"])
                for x in [x for x in updates if "edited_message" in x]:
                    self.http.messages_dict[x["edited_message"]["message_id"]] = Message(self.http, x["edited_message"])

            except InvalidToken as exc:
                traceback.print_exception(type(exc), exc, exc.__traceback__, file=sys.stderr)
                await self.stop()
                return
            except HTTPException as exc:
                traceback.print_exception(type(exc), exc, exc.__traceback__, file=sys.stderr)
                await asyncio.sleep(10)
            except Exception as exc:
                traceback.print_exception(type(exc), exc, exc.__traceback__, file=sys.stderr)
                await asyncio.sleep(10)

            log.debug(f"Waiting for {self._wait_time} seconds")
            await asyncio.sleep(self._wait_time)

    async def _use_event_handler(self, func, *args, **kwargs):
        try:
            await func(*args, **kwargs)
        except Exception as exc:
            await self._dispatch("error", exc)

    async def _dispatch(self, event, *args):
        log.debug(f"Dispatchng {event} with {args}")
        # Handle the active wait_fors
        waiting_for = self._waiting_for.get(event)
        if waiting_for:
            removed_futures = []
            for i, (future, check) in enumerate(waiting_for):
                if future.cancelled():
                    removed_futures.append(i)
                    continue

                # Run the check associated with the future
                # and set the future's result or exception
                try:
                    result = check(*args)
                except Exception as exc:
                    future.set_exception(exc)
                    removed_futures.append(i)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed_futures.append(i)

            # Clean up waiting_for
            if len(removed_futures) == len(waiting_for):
                self._waiting_for.pop(event)
            else:
                for idx in reversed(removed_futures):
                    del waiting_for[idx]

        # Add "on_" to the event name
        event = f"on_{event}"

        # Get the listeners for the event
        if event in self._listeners:
            handlers = self._listeners[event]
        else:
            handlers = []

        try:
            handlers.append(getattr(self, event))
        except AttributeError:
            pass

        # Dispatch the event to the listeners
        for handler in handlers:
            self.loop.create_task(self._use_event_handler(handler, *args))

    async def wait_for(self, event: str, check=None, timeout=None):
        """|coro|

        Waits for an event.

        Parameters
        ----------
        event: :class:`str`
            The name of the event to wait for.
        """
        ev = event.lower()

        if not check:
            def check(*args):
                return True

        future = self.loop.create_future()

        try:
            waiting_for = self._waiting_for[ev]
        except KeyError:
            waiting_for = []
            self._waiting_for[ev] = waiting_for

        waiting_for.append((future, check))

        return await asyncio.wait_for(future, timeout=timeout)

    def event(self, func):
        """
        Turns a function into an event handler.

        Parameters
        ----------
        func:
            The function to make an event handler.
        """

        setattr(self, func.__name__, func)
        return func

    async def on_error(self, error):
        """|coro|

        Default error handler.
        """

        if "on_error" in self._listeners:
            return

        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    def add_listener(self, func, name: str = None):
        """
        Registers a function as a listener.

        Parameters
        ----------
        func:
            The function to register.
        name: Optional[:class:`str`]
            The name of the event to register the function as.
        """

        name = name or func.__name__

        if name in self._listeners:
            self._listeners[name].append(func)
        else:
            self._listeners[name] = [func]

    def remove_listener(self, func):
        """
        Removes a listener.

        Parameters
        ----------
        func:
            The function that is registered as a listener.
        """

        for event in self._listeners:
            if func in self._listeners[event]:
                self._listeners[event].remove(func)

    def listen(self, name=None):
        """
        A decorator that registers a function as a listener.

        Parameters
        ---------
        name: Optional[:class:`str`]
             The name of the event to register the function as.
        """

        def deco(func):
            self.add_listener(func, name)
            return func

        return deco

    async def start(self):
        """|coro|
        Starts the bot.
        """

        if self._running:
            raise RuntimeError("Bot is already running")

        self._running = True
        self.loop.create_task(await self._poll())

    async def stop(self):
        """|coro|
        Stops the bot.
        """

        if not self._running:
            raise RuntimeError("Bot is not running")

        log.info("Stopping the bot")
        await self.http.close()
        self._running = False

    def _clean_tasks(self):
        pending = asyncio.all_tasks(loop=self.loop)
        if not pending:
            return

        for task in pending:
            task.cancel()
        self.loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))

    def run(self):
        """
        Runs the bot.
        """

        self._running = True

        try:
            log.info("Running the bot")
            self.loop.run_until_complete(self._poll())
        except KeyboardInterrupt:
            log.info("Received KeyboardInterrupt, Stopping bot")
            self.loop.run_until_complete(self.stop())
        finally:
            self._clean_tasks()
            self.loop.run_until_complete(self.loop.shutdown_asyncgens())
            self.loop.close()



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

import html
import itertools
import typing

import Octrapy

from .core import Command
from .cog import Cog


class _HelpCommandImplementation(Command):
    """Class that interfaces with :class:`Octrapy.ext.commands.Command`."""

    def __init__(self, help_cmd, bot, command_attrs):
        self.help_cmd = help_cmd

        super().__init__(help_cmd, **command_attrs)
        self.bot = bot

class HelpCommand:
    """Help command template.

    Attributes
    ----------
    ctx: :class:`Octrapy.ext.commands.Context`
        The :class:`Octrapy.ext.commands.Context` for the command
    bot: :class:`Octrapy.ext.commands.Bot`
        The :class:`Octrapy.ext.commands.Bot` from the Context
    """

    def __init__(self, **options):
        self.command_attrs = options.pop('command_attrs', {})
        self.command_attrs.setdefault("name", "help")
        self.command_attrs.setdefault("description", "The help command")
        self.command_attrs.setdefault("aliases", ["start"])

        self._implementation = None

    def _add_to_bot(self, bot):
        implementation = _HelpCommandImplementation(self, bot, self.command_attrs)
        bot.add_command(implementation)
        self._implementation = implementation

    def _remove_from_bot(self, bot):
        bot.remove_command(self._implementation.name)
        self._implementation = None

    async def get_command_signature(self, command):
        """The method that gets a formatted command signature

        Example:
        /help [command]
        """
        name = html.escape(command.name)
        sig = html.escape(command.signature)
        return f"/{name} {sig}"

    async def send_bot_help(self):
        """The method that sends help for the bot.

        This is called when no query is provided.
        This method should handle the sending of the help message.
        """
        raise NotImplementedError("Subclasses must implement this.")

    async def send_cog_help(self, cog: Cog):
        """The method that sends help for a cog.

        This is called when a cog matches the query.
        This method should handle the sending of the help message.

        Parameters
        ----------
        cog: :class:`Octrapy.ext.commands.Cog`
            The cog that matched the query
        """
        raise NotImplementedError("Subclasses must implement this.")

    async def send_command_help(self, command: Command):
        """The method that sends help for a command.

        This is called when a command matches the query.
        This method should handle the sending of the help message.

        Parameters
        ----------
        command: :class:`Octrapy.ext.commands.Command`
            The command that matched the query
        """
        raise NotImplementedError("Subclasses must implement this.")

    async def send_not_found(self, query: str):
        """The method that sends a 'not found' message or similar.

        This method is called when no match is found for the query.

        Parameters
        ----------
        query: :class:`str`
            The user's query
        """
        await self.ctx.send(f"A command or cog named '{query}' was not found.")

    async def help_callback(self, query: typing.Optional[str]):
        """The callback that searches for a matching commmand or cog.

        This should not be overridden unless it is necessary.

        Parameters
        ----------
        query: Optional[:class:`str`]
            The user's query. Defaults to ``None``.
        """
        bot = self.bot

        # Send the bot help if there is no query
        if query is None:
            await self.send_bot_help()
            return

        # Check if the query matches a cog
        cogs = bot.cogs

        if query in cogs.keys():
            cog = cogs[query]
            await self.send_cog_help(cog)
            return

        # If not, check if the query matches a command
        command = bot.get_command(query)
        if command:
            await self.send_command_help(command)
            return

        # If neither, send the not found message
        await self.send_not_found(query)

    async def __call__(self, ctx, *, command=None):
        self.ctx = ctx
        self.bot = ctx.bot
        await self.help_callback(command)


class DefaultHelpCommand(HelpCommand):
    """The default help command.

    This help command mimics BotFather's help command look.

    Parameters
    ----------
    no_category: Optional[:class:`str`]
        The heading for commands without a category.
        Defaults to "No Category".
    sort_commands: Optional[:class:`bool`]
        Whether to sort the commands.
        Defaults to ``True``.
    """

    def __init__(self, **options):
        self.no_category = options.pop("no_category", "No Category")
        self.sort_commands = options.pop("sort_commands", True)
        super().__init__(**options)

    def get_ending_note(self):
        """Returns the command's ending note."""
        name = self._implementation.name
        return (
            f"Type /{name} [command] for more info on a command.\n"
            f"You can also type /{name} [category] for more info on a category."
        )

    async def format_commands(self, commands: typing.List[Command], *, heading: str):
        """The method that formats a given list of commands.

        Parameters
        ----------
        commands: List[:class`Octrapy.ext.commands.Command`]
            The list of commands to format.
        heading: :class:`str`
            The heading to display.
        """
        if not commands:
            return []

        formatted = []

        formatted.append(f"<b>{html.escape(heading)}:</b>")

        def make_entry(sig, doc, *, alias_for=None):
            alias = f"[Alias for {alias_for}] " if alias_for else ""

            if doc:
                return f"{sig} - {alias}{html.escape(doc)}"
            else:
                entry = f"{sig}"
                if alias:
                    entry += f" {alias}"
                return entry

        for command in commands:
            if command.hidden:
                continue

            sig = await self.get_command_signature(command)
            doc = command.description
            formatted.append(make_entry(sig, doc))

        return formatted

    async def format_command(self, command):
        """The method that formats an indivitual command.

        Parameters
        ------------
        command: :class:`Command`
            The command to format.
        """

        help_text = [await self.get_command_signature(command)]

        if command.description:
            help_text.append(html.escape(command.description))
        if command.aliases:
            help_text.append(f"Aliases: {', '.join(command.aliases)}")

        return help_text

    async def send_help_text(self, help_text):
        message = "\n".join(help_text)
        await self.ctx.send(message, parse_mode="HTML")

    async def send_bot_help(self):
        bot = self.bot

        help_text = []

        if bot.description:
            # <description> portion
            help_text.append(html.escape(bot.description))
            help_text.append("")  # blank line

        no_category = self.no_category

        def get_category(command, *, no_category=no_category):
            cog = command.cog
            return cog.qualified_name if cog is not None else no_category

        to_iterate = itertools.groupby(bot.commands, key=get_category)

        # Now we can add the commands to the page.
        for category, commands in to_iterate:
            commands = (
                sorted(commands, key=lambda c: c.name)
                if self.sort_commands
                else list(commands)
            )
            added = await self.format_commands(commands, heading=category)
            if added:
                help_text.extend(added)
                help_text.append("")  # blank line

        note = self.get_ending_note()
        if note:
            # help_text.append("")  # blank line
            help_text.append(html.escape(note))

        await self.send_help_text(help_text)

    async def send_cog_help(self, cog: Cog):
        bot = self.bot

        help_text = []

        if cog.description:
            help_text.append(html.escape(cog.description))
            help_text.append("")  # blank line

        help_text.extend(await self.format_commands(cog.commands, heading="Commands"))

        note = self.get_ending_note()
        if note:
            help_text.append("")  # blank line
            help_text.append(html.escape(note))

        await self.send_help_text(help_text)

    async def send_command_help(self, command: Command):
        await self.send_help_text(await self.format_command(command))



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

from .file import *
from .abc import OctraObject


class User(OctraObject):
    """
    Represents a Octra user.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: x != y

            Checks if two users are not equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    id: :class:`int`
        The ID of the user.
    is_bot: :class:`bool`
        If the user is a bot.
    username: Optional[:class:`str`]
        The username of the user.
    first_name: :class:`str`
        The first name of the user.
    last_name: Optional[:class:`str`]
        The last name of the user.
    """

    def __init__(self, http, data):
        super().__init__(http, data)
        self.is_bot = data.get("is_bot")
        self.username = data.get("username")
        self.first_name = data.get("first_name")
        self.last_name = data.get("last_name")

    def __str__(self):
        return self.username

    @property
    def full_name(self):
        """
        :class:`str`:
             The user's full name.
        """

        return f"{self.first_name or ''} {self.last_name or ''}"

    @property
    def name(self):
        """
        :class:`str`:
            Username if the user has one. Otherwise the full name of the user.
        """

        return self.username or self.full_name

    async def send(self, content: str = None, file: File = None, parse_mode: str = None):
        """|coro|
        
        Sends a message directly to the user.

        Parameters
        ----------
        content: :class:`str`
            The content of the message to send.
        file: :class:`Octrapy.File`
            The file to send
        parse_mode: :class:`str`
            The parse mode of the message to send.

        Returns
        -------
        :class:`Octrapy.Message`
            The message sent.

        Raises
        ------
        :exc:`errors.HTTPException`
            Sending the message failed.
        """

        if not file:
            return await self._http.send_message(chat_id=self.id, content=content, parse_mode=parse_mode)
        else:
            if isinstance(file, Document):
                return await self._http.send_document(chat_id=self.id, document=file.file, filename=file.filename)

            elif isinstance(file, Photo):
                return await self._http.send_photo(chat_id=self.id, photo=file.file, filename=file.filename, caption=file.caption)



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

from .errors import BadArgument
from Octrapy.chat import Chat
from Octrapy.user import User
from Octrapy.errors import HTTPException

class Converter:
    """
    Base class for converters.
    """

    async def convert(self, ctx, arg):
        """
        Does the converting.
        """
        raise NotImplementedError

class UserConverter(Converter):
    """
    Converts an argument into a user.
    """

    async def convert(self, ctx, arg):
        try:
            arg = int(arg)
        except ValueError:
            raise BadArgument(arg, User, "Argument is not an ID")
        try:
            return await ctx.chat.get_member(arg)
        except HTTPException:
            raise BadArgument(arg, User, "Failed to fetch user")

class ChatConverter(Converter):
    """
    Converts an argument into a chat.
    """

    async def convert(self, ctx, arg):
        try:
            arg = int(arg)
        except ValueError:
            raise BadArgument(arg, Chat, "Argument is not an ID")
        try:
            return await ctx.bot.get_chat(arg)
        except HTTPException:
            raise BadArgument(arg, Chat, "Failed to fetch chat")



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

import datetime

from .chat import Chat
from .user import User
from .abc import OctraObject


class Message(OctraObject):
    """
    Represents a message in Octra.

    .. container:: operations

        .. describe:: x == y

            Checks if two messages are equal.

        .. describe:: x != y

            Checks if two messages are not equal.

    Attributes
    ----------
    id: :class:`int`
        The ID of the message.
    created_at: :class:`datetime.datetime`
        The time the message was created.
    edited_at: Optional[:class:`datetime.datetime`]
        The time the message was edited.
    content: :class:`str`
        The content of the message.
    chat: :class:`Octrapy.Chat`
        The chat the message is in.
    author: :class:`Octrapy.User`
        The author of the message.
    """

    def __init__(self, http, data: dict):
        super().__init__(http, data)
        self.id = data.get("message_id")

        self.created_at = data.get("date")
        if self.created_at:
            datetime.datetime.fromtimestamp(self.created_at)

        self.edited_at = data.get("edit_date")
        if self.edited_at:
            datetime.datetime.fromtimestamp(self.edited_at)

        self.content = data.get("text")

        if "chat" in data:
            self.chat = Chat(http, data.get("chat"))
        else:
            self.chat = None

        if "from" in data:
            self.author = User(http, data.get("from"))
        else:
            self.author = None

    async def reply(self, content: str, parse_mode: str = None):
        """|coro|
        
        Replys to the message.

        Parameters
        ----------
        content: :class:`str`
            The content of the message to send.
        parse_mode: :class:`str`
            The parse mode of the message to send.

        Returns
        -------
        :class:`Octrapy.Message`
            The message sent.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Sending the message failed.
        """

        return await self._http.send_message(chat_id=self.chat.id, content=content, parse_mode=parse_mode, reply_message_id=self.id)

    async def forward(self, destination):
        """|coro|
        
        Forwards the message to a destination.

        Parameters
        ----------
        destination: :class:Octrapy.Chat`
            The chat forward the message to.
        
        Returns
        -------
        :class:`Octrapy.Message`
            The message sent.
        
        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Forwarding the message failed.
        """

        return await self._http.forward_message(chat_id=destination.id, from_chat_id=self.chat.id, message_id=self.id)

    async def edit(self, content: str, parse_mode: str = None):
        """|coro|

        Edits the message.

        Parameters
        ----------
        content: :class:`str`
            The content of the new message.
        parse_mode: :class:`str`
            The parse mode of the new message.

        Raises
        ------
        :exc:`Octrapy.HTTException`
            Editing the message failed.
        """

        await self._http.edit_message_content(chat_id=self.chat.id, message_id=self.id, content=content, parse_mode=parse_mode)

    async def delete(self):
        """|coro|

        Deletes the message.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Deleting the message failed.
        """

        return await self._http.delete_message(chat_id=self.chat.id, message_id=self.id)



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

class OctraObject:
    """Base Octra object."""

    def __init__(self, http, data):
        self._http = http
        self._data = data

        self.id = data.get("id")

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id  == other.id

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return other.id != self.id
        return True



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

import re


def escape_markdown(text: str, *, version: int = 2):
    """Tool that escapes markdown from a given string.

    Parameters
    ----------
    text: :class:`str`
        The text to escape markdown from.
    version: Optional[:class:`int`]
        The Octra markdown version to use. Only 1 and 2 are supported.

    Returns
    -------
    :class:`str`
        The escaped text.

    Raises
    ------
    :exc:`ValueError`
        An unsupported version was provided.
    """
    if version == 1:
        characters = r"_*`["

    elif version == 2:
        characters = r"_*[]()~`>#+-=|{}.!"

    else:
        raise ValueError(f"Version '{version}' unsupported. Only version 1 and 2 are supported.")

    return re.sub(f"([{re.escape(characters)}])", r"\\\1", text)



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

from .errors import *
from .file import *
from .abc import OctraObject


class Chat(OctraObject):
    """
    Represents a chat in Octra.

    .. container:: operations

        .. describe:: x == y

            Checks if two chats are equal.

        .. describe:: x != y

            Checks if two chats are not equal.

        .. describe:: str(x)

            Returns the chat's title.


    Attributes
    ----------
    id: :class:`int`
        The ID of the chat.
    title: :class:`str`
        The title of the chat.
    description: Optional[:class:`str`]
        The description of the chat.
    type: :class:`str`
        The type of the chat.
    """

    def __init__(self, http, data: dict):
        super().__init__(http, data)
        self.title = data.get("title")
        self.username = data.get("username")
        self.description = data.get("description")
        self.type = data.get("type")

    def __str__(self):
        return self.title

    async def send(self, content: str = None, file: File = None, parse_mode: str = None):
        """|coro|

        Sends a message to the chat.

        Parameters
        ----------
        content: :class:`str`
            The content of the message to send.
        file: :class:`Octrapy.File`
            The file to send.
        parse_mode: :class:`str`
            The parse mode of the message to send.

        Returns
        -------
        :class:`Octrapy.Message`
            The message sent.

        Raises
        ------
        :exc:`errors.HTTPException`
            Sending the message failed.
        """

        if not file:
            return await self._http.send_message(chat_id=self.id, content=content, parse_mode=parse_mode)
        else:
            if isinstance(file, Document):
                return await self._http.send_document(chat_id=self.id, document=file.file, filename=file.filename)

            elif isinstance(file, Photo):
                return await self._http.send_photo(chat_id=self.id, photo=file.file, filename=file.filename, caption=file.caption)

    async def send_poll(self, question: str, options: list):
        """|coro|

        Sends a poll to the chat.

        Parameters
        ----------
        question: :class:`str`
            The question of the poll.
        options: :class:`list`
            The options in the poll.

        Returns
        -------
        :class:`Octrapy.Poll`
            The poll sent.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Sending the poll failed.
        """

        return await self._http.send_poll(chat_id=self.id, question=question, options=options)

    async def send_action(self, action: str):
        """|coro|

        Sends an action to the chat.

        Parameters
        ----------
        action: :class:`str`
            The action to send.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Sending the action failed.
        """

        await self._http.send_chat_action(chat_id=self.id, action=action)

    async def get_member(self, user_id: int):
        """|coro|

        Fetches a member in the chat.

        Parameters
        ----------
        user_id: :class:`int`
            The user ID of the member.

        Returns
        -------
        :class:`Octrapy.User`
            The member fetched.

        Raises
        ------
        :exc:`Octrapy.HTTPException`
            Fetching the member failed.
        """

        return await self._http.get_chat_member(chat_id=self.id, user_id=user_id)

    @property
    def history(self):
        """
        :class:`list`:
            The cached messages in the chat.
        """

        return [x for x in self._http.messages if x.chat.id == self.id]

    def fetch_message(self, message_id: int):
        """
        Fetches a message from the cache.

        Parameters
        ----------
        message_id: :class:`int`
            The ID of the message to fetch.

        Returns
        -------
        :class:`chat.Message`
            The message fetched.
        """

        for x in self.history:
            if x.id == message_id:
                return x



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

import inspect

import Octrapy
from Octrapy import User, Chat

from .errors import *
from .converter import *
from .context import Context

class Command:
    """
    Represents a command.

    Attributes
    ----------
    name: :class:`str`
        The name of the command.
    description: :class:`str`
        The description of the command.
    usage: :class:`str`
        The usage of the command.
    aliases: :class:`str`
        The aliases for the command.
    callback:
        The callback of the command.
    hidden: :class:`bool`
        If the command is hidden.
    cog: :class:`Octrapy.ext.commands.Cog`
        The cog the command is in.
    bot: :class:`Octrapy.ext.commands.Bot`
        The bot the command is in.
    """

    def __init__(self, func, **kwargs):
        self.callback = func

        signature = inspect.signature(func)
        self.params = signature.parameters.copy()

        self._data = kwargs
        self.name = kwargs.get("name") or func.__name__
        self.description = kwargs.get("description")
        self.usage = kwargs.get("usage")
        self.aliases = kwargs.get("aliases") or []
        self.hidden = kwargs.get("hidden") or False
        self.cog = None
        self.bot = None
        self.checks = []

    def __str__(self):
        return self.name

    @property
    def clean_params(self):
        """OrderedDict[:class:`str`, :class:`inspect.Parameter`]:
        Returns a mapping similar to :attr:`inspect.Signature.parameters`,
        but without self or context.
        """
        params = self.params.copy()

        # if in a cog, the first parameter is self
        if self.cog is not None:
            params.popitem(last=False)

        # the context parameter is always first/next
        try:
            params.popitem(last=False)
        except Exception:
            raise ValueError("Missing context parameter") from None

        return params

    @property
    def signature(self):
        """:class:`str`: Returns a signature for a command that can be used in help commands

        <required> required param
        [optional] optional param
        [params...] optional list of params
        [optional=0] optional param defaults to 0
        """
        if self.usage:
            return self.usage

        params = self.clean_params

        if not params:
            return ""

        final = []
        for name, param in params.items():
            if param.default is not param.empty:
                if param.default is None or (isinstance(param.default, str) and not param.default):
                    final.append(f"[{name}]")

                else:
                    final.append(f"[{name}={param.default}]")

            elif param.kind == param.VAR_POSITIONAL:
                final.append(f"[{name}...]")

            else:
                final.append(f"<{name}>")

        return " ".join(final)

    def add_check(self, func):
        """
        Adds a check.

        Parameters
        ----------
        func:
            The function to add to the checks.
        """

        self.checks.append(func)

    def remove_check(self, func):
        """
        Removes a check.

        Parameters
        ----------
        func:
           The function to remove from the checks.
        """

        if func not in self.checks:
            return

        self.checks.remove(func)

    async def _convert_arg(self, ctx, typehint, arg):
        # Go through the possible Octra objects and try to find the needed converter
        if typehint == User:
            converter = UserConverter()
        elif typehint == Chat:
            converter = ChatConverter()

        # Otherwise just set the converter to the typehing
        else:
            converter = typehint

        # Attempt to convert the arg
        try:
            if isinstance(converter, Converter):
                return await converter.convert(ctx, arg)
            else:
                return converter(arg)
        except Exception as exc:
            # If the error is already BadArgument just re-raise the error
            if isinstance(exc, BadArgument):
                raise
            # Otherwise take the converter and given argument and raise a generic BadArgument error
            raise BadArgument(arg, typehint.__name__) from None


    async def _parse_args(self, ctx: Context):
        given_args = ctx.message.content.split(" ")
        given_args.pop(0)

        if ctx.command:
            takes_args = [
                x[1]
                for x in list(
                    inspect.signature(ctx.command.callback).parameters.items()
                )
            ]
            if ctx.command.cog:
                takes_args.pop(0)
            takes_args.pop(0)

            # Iter through the arguments
            for counter, argument in enumerate(takes_args):
                try:
                    # If argument can be positional, give one arg
                    if argument.kind in (
                        inspect._ParameterKind.POSITIONAL_OR_KEYWORD,
                        inspect._ParameterKind.POSITIONAL_ONLY,
                    ):
                        give = given_args[0]
                        converter = argument.annotation
                        # If the argument as a converter, try and convert
                        if converter != inspect._empty:
                            value = await self._convert_arg(ctx, converter, give)
                        else:
                            value = give
                        ctx.args.append(value)
                        given_args.pop(0)

                    # If argument is a keyword argument, give the rest of the arguments
                    elif argument.kind == inspect._ParameterKind.KEYWORD_ONLY:
                        give = " ".join(given_args)
                        if give == "":
                            raise IndexError()

                        converter = argument.annotation
                        # If the argument has a converter, try and convert
                        if converter != inspect._empty:
                            value = await self._convert_arg(ctx, converter, give)
                        else:
                            value = give
                        ctx.kwargs[argument.name] = value

                    elif argument.kind == inspect._ParameterKind.VAR_POSITIONAL:
                        if len(given_args) == 0:
                            raise IndexError()
                        for give in given_args:
                            converter = argument.annotation
                            if converter != inspect._empty:
                                value = await self._convert_arg(ctx, converter, give)
                            else:
                                value = give
                            ctx.args.append(value)

                except IndexError:
                    # If no argument does not have a default, raise MissingRequiredArgument
                    if argument.default == inspect._empty:
                        raise MissingRequiredArgument(argument.name) from None
                    # Otherwise set the argument to the default
                    if argument.kind in (
                        inspect._ParameterKind.POSITIONAL_OR_KEYWORD,
                        inspect._ParameterKind.POSITIONAL_ONLY,
                    ):
                        ctx.args.append(argument.default)

    async def invoke(self, ctx: Context):
        """
        |coro|
        
        Invokes the command with given context.

        Parameters
        ----------
        ctx: :class:`Octrapy.ext.commands.Context`
            The context to invoke the command with.
        """

        # Checks
        for check in self.checks:
            if not check(ctx):
                raise CheckFailure("A check for this command has failed")

        if self.cog:
            if not self.cog.cog_check(ctx):
                raise CheckFailure("A check for this command has failed")

        other_args = []
        if self.cog:
            other_args.append(self.cog)
        other_args.append(ctx)

        await self._parse_args(ctx)

        try:
            return await self.callback(*other_args, *ctx.args, **ctx.kwargs)
        except Exception as exc:
            raise CommandInvokeError(exc) from exc

def command(*args, **kwargs):
    """Turns a function into a command.

    See :class:`Octrapy.ext.commands.Command`
    for parameters.
    """

    def deco(func):
        kwargs["name"] = kwargs.get("name")

        command = Command(func, **kwargs)
        command.checks = getattr(func, "_command_checks", [])

        command = Command(func, **kwargs)
        return command

    return deco


def check(check_function):
    """Makes a check for a command."""

    def deco(func):
        if isinstance(func, Command):
            func.add_check(check_function)
        else:
            if not getattr(func, "_command_checks", None):
                func._command_checks = [check_function]
            else:
                func._command_checks.append(check_function)
        return func

        return

    return deco


def is_owner():
    """A command check for checking that the user is the owner."""

    def is_owner_check(ctx):
        if ctx.author.id not in (ctx.bot.owner_ids or [ctx.bot.owner_id]):
            raise NotOwner("You must be the owner to use this command")
        return True

    return check(is_owner_check)


def is_private_chat():
    """A command check for checking that the chat is a private chat."""

    def is_private_chat_check(ctx):
        if ctx.chat.type != "private":
            raise PrivateChatOnly()
        return True

    return check(is_private_chat_check)


def is_not_private_chat():
    """A command check for checking that the chat is not a private chat."""

    def is_not_private_chat_check(ctx):
        if ctx.chat.type == "private":
            raise GroupOnly()
        return True

    return check(is_not_private_chat_check)



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

from io import BytesIO


class File:
    """A file to send."""

    def __init__(self, file: BytesIO, filename: str = None):
        self.file = file
        self.filename = filename


class Document(File):
    """
    A document to send.

    Attributes
    ----------
    file: :class:`io.BytesIO`
        The document.
    filename: Optional[:class:`str`]
        The filename of the document.

    Parameters
    ----------
    file: :class:`io.BytesIO`
        The document.
    filename: :class:`str`
        The filename of the document.
    """


class Photo(File):
    """
    A photo to send.

    Attributes
    ----------
    file: :class:`io.BytesIO`
        The photo.
    filename: Optional[:class:`str`]
        The filename of the photo.
    caption: Optional[:class:`str`]
        The caption of the photo.

    Parameters
    ----------
    file: :class:`io.BytesIO`
        The photo.
    filename: :class:`str`
        The filename of the photo.
    caption: :class:`str`
        The caption of the photo.
    """

    def __init__(self, file: BytesIO, filename: str = None, caption: str = None):
        super().__init__(file, filename)
        self.caption = caption



import setuptools
import re

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = []
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

version = ""
with open("Octrapy/__init__.py") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(1)

setuptools.setup(
    name="Octra.py",
    description="An async API wrapper for Octra bot API in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Akash",
    url="https://github.com/Octra/Octra.py",
    project_urls={
        "Documentation": "https://Octrapy.readthedocs.io",
        "Issue tracker": "https://github.com/Octra/Octra.py/issues",
    },
    version=version,
    packages=["Octrapy", "Octrapy/ext/commands"],
    install_requires=requirements,
    extras_require={
        "docs": [
            "sphinx==2.4.3",
            "sphinx-rtd-theme",
            "sphinxcontrib_trio==1.1.1",
            "sphinxcontrib-websupport",
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

class Poll:
    """A Octra poll.

    Attributes
    ----------
    question: :class:`str`
        The question of the poll.
    options: :class:`list`
        The options of the poll.
    total_voter_count: :class:`int`
        The total voter count of the poll.
    is_closed: :class:`bool`
        If the poll is closed.
    is_anonymous: :class:`bool`
        If the poll is anonymous.
    type: :class:`str`
        The type of the poll.
    allow_multiple_answers: :class:`bool`
        If the poll allows multiple answers.
    """

    def __init__(self, data):
        self._data = data

        self.id = data.get("id")

        self.question = data.get("question")
        self.options = data.get("options")
        self.total_voter_count = data.get("total_voter_count")
        self.is_closed = data.get("is_closed")
        self.is_anonymous = data.get("is_anoymous")
        self.type = data.get("type")
        self.allow_multiple_answers = data.get("allow_multiple_answers")



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

import types
import inspect

from .core import Command


class CogMeta(type):
    """Metaclass for Cog."""

    def __new__(cls, *args, **kwargs):
        name, bases, attrs = args
        attrs["__cog_name__"] = kwargs.pop("name", name)

        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)
        commands = []
        listeners = []

        for base in new_cls.__mro__:
            for command in base.__dict__.values():
                # Add the command if object is a command
                if isinstance(command, Command):
                    commands.append(command)

                # If object is a method and it has _cog_listener attribute, add the listener
                elif isinstance(command, types.FunctionType):
                    if hasattr(command, "_cog_listener"):
                        try:
                            listeners.append(command)
                        except AttributeError:
                            pass

        new_cls.commands = commands
        new_cls.listeners = listeners
        return new_cls


class Cog(metaclass=CogMeta):
    """
    Base cog class.

    Attributes
    ----------
    commands: List[:class:`Octrapy.Command`]
        The cog's commands.
    listeners: :class:`list`
        The cog's listeners.
    """

    @property
    def qualified_name(self):
        """:class:`str`: The cog's name."""
        return self.__cog_name__

    @property
    def description(self):
        """:class:`str`: The cog's description."""
        return inspect.getdoc(self)

    @classmethod
    def listener(cls, name: str = None):
        """
        Makes a method in a cog a listener.

        Parameters
        ----------
        name: Optional[:class:`str`]
            The name of the event to register the function as.
        """

        def deco(func):
            func._cog_listener = name or func.__name__
            return func

        return deco
    
    def _add(self, bot):
        for command in self.commands:
            command.bot = bot
            command.cog = self
            bot.add_command(command)
        for listener in self.listeners:
            bot.add_listener(getattr(self, listener.__name__), listener._cog_listener)

    def _remove(self, bot):
        for command in self.commands:
            bot.remove_command(command.name)
        for listener in self.listeners:
            bot.remove_listener(getattr(self, listener.__name__))

    def cog_check(self, ctx):
        return True



"""
MIT License

Copyright (c) 2023 Akash

Octra ltd
"""

from Octrapy import OctraException


class CommandError(OctraException):
    """Base exception for all command errors."""

    pass

class CommandNotFound(CommandError):
    """Raised when a command is not found."""

    pass

class CommandRegistrationError(CommandError):
    """Raised when a command cannot be registered."""

    pass

class ExtensionNotLoaded(CommandError):
    """
    Raised when an extension is not loaded.

    Attributes
    ----------
    name: :class:`str`
        The name of the extension that is not loaded.
    """

    def __init__(self, name):
        self.name = name
        super().__init__(f"Extension '{name}' has not been loaded")

class ExtensionAlreadyLoaded(CommandError):
    """
    Raised when an extension is already loaded.

    Attributes
    ----------
    name: :class:`str`
        The name of the extension that is already loaded.
    """

    def __init__(self, name):
        self.name = name
        super().__init__(f"Extension '{name}' is already loaded")

class MissingRequiredArgument(CommandError):
    """
    Raised when a required argument is missing.

    Attributes
    ----------
    param: :class:`str`
        The argument that is missing.
    """

    def __init__(self, param):
        self.param = param
        super().__init__(f"'{param}' is a required argument that is missing")

class BadArgument(CommandError):
    """
    Raised when a bad argument is given.

    Attributes
    ----------
    arg: :class:`str`
        The bad argument.
    converter: :class:`str`
        The name of the converter that failed.
    """

    def __init__(self, arg, converter, message=None):
        self.arg = arg
        self.converter = converter
        super().__init__(message or f"Failed to convert '{arg}' to '{converter}'")

class CheckFailure(CommandError):
    """Raised when a check fails."""

    pass

class NotOwner(CheckFailure):
    """Raised when a user is not the owner of the bot."""

    pass

class CommandInvokeError(CommandError):
    """
    Raised when a command fails.

    Attributes
    ---------
    error: :class:`Exception`
         The original error that was raised.
    """

    def __init__(self, error):
        self.error = error
        super().__init__(f"Command raised an exception: {error.__class__.__name__}: {error}")

class PrivateChatOnly(CheckFailure):
    """Raised when a command can only be used in private chats."""

    def __init__(self, message=None):
        super().__init__(message or "This command can only be used in private messages")

class GroupOnly(CheckFailure):
    """Raised when a command can only be used in groups."""

    def __init__(self, message=None):
        super().__init__(message or "This command can only be used in groups")

    pass
